"""
Created on Wed Apr 19 21:39:45 2023

@author: Alexander Stum
GIS Specialist
National Soil Survey Center
USDA-NRCS
e-mail: alexander.stum@usda.gov

@author: Adolfo.Diaz
         GIS Specialist
         Soil Services and Information
         USDA - NRCS
e-mail: adolfo.diaz@usda.gov

"""
import arcpy 
import sys
import os
import traceback
import shutil
import zipfile
import socket
import requests
import io
import logging
from datetime import datetime
from time import sleep
from urllib.request import URLError, HTTPError
from arcpy import env
import multiprocessing as mp
import concurrent.futures as cf
import itertools as it


# def poolStatus(ssa):
#     areaSym, outcome = ssa.items()
#     if not outcome:
#         failedCnt += 1



def arcpyErr(func):
    try:
        etype, exc, tb = sys.exc_info()
        line = tb.tb_lineno
        msgs = f"ArcPy ERRORS:\nIn function: {func} on line: {line}\n{arcpy.GetMessages(2)}\n"
        arcpy.AddError(msgs)
        return msgs
    except:
        arcpy.AddError("Error in arcpyErr method")

        
def pyErr(func):
    try:
        etype, exc, tb = sys.exc_info()
        
        tbinfo = traceback.format_tb(tb)[0]
        msgs = f"PYTHON ERRORS:\nTraceback info:\nIn function: {func}\n{tbinfo}\nError Info:\n{exc}"
        arcpy.AddError(msgs)
        return msgs
    except:
        arcpy.AddError("Error in pyErr method")

def concurrently(fn, max_concurrency, iterSets, constSets ):
    """
    Adapted from https://github.com/alexwlchan/concurrently/blob/main/concurrently.py
    

    Generates (input, output) tuples as the calls to ``fn`` complete.

    See https://alexwlchan.net/2019/10/adventures-with-concurrent-futures/ for an explanation
    of how this function works.
    
    Parameters
    ----------
    fn : function
        The function that it to be run in parallel.
        
    max_concurrency : int
        Maximum number of processes, parameter for itertools islice function.
        
    iterSets : list
        List of dictionaries that will be iterated through. The keys of the dictionary
        must be the same for each dictionary and align with keywords of the function.
        
    constSets : dict
        Dictionary of parameters that constant for each iteration of the function
        ``fn``. Dictionary keys must align with function keywords.
        

    """
    try:
        # Make sure we get a consistent iterator throughout, rather than
        # getting the first element repeatedly.
        # count = len(iterSets)
        fn_inputs = iter(iterSets)
    
        with cf.ThreadPoolExecutor() as executor:
            # initialize first set of processes
            futures = {
                executor.submit(fn, **params, **constSets): params
                for params in it.islice(fn_inputs, max_concurrency)
            }
            # Wait for a future to complete, returns sets of complete and incomplete futures
            while futures:
                done, _ = cf.wait(
                    futures, return_when = cf.FIRST_COMPLETED
                )
    
                for fut in done:
                    # once process is done clear it out, yield results and params
                    original_input = futures.pop(fut)
                    yield original_input, fut.result()
                
                # Sends another set of processes equivalent in size to those just completed
                # to executor to keep it at max_concurrency in the pool at a time,
                # to keep memory consumption down.
                futures.update({executor.submit(fn, **params, **constSets): params
                for params in it.islice(fn_inputs, len(done))
                })
    except GeneratorExit:
        raise
        # arcpy.AddError("Need to do some clean up.")
        # yield 2, 'yup'
    except arcpy.ExecuteError:
        func = sys._getframe(  ).f_code.co_name
        msgs = arcpyErr(func)
        yield [2, msgs]
    except:
        func = sys._getframe(  ).f_code.co_name
        msgs = pyErr(func)
        yield [2, msgs]
## ===================================================================================
def CheckExistingDataset(areaSym, surveyDate, newFolder):
    """Checks if a most current and complete download of the survey exist
    

    Parameters
    ----------
    areaSym : str
        The area symbol of current soil survey area being processed.
    surveyDate : str
        The date the soil survey area was updated on WSS.
    newFolder : str
        Path of the soil survey download.
    newDB : str
        Path for the soil survey Template, None if option not selected.

    Returns
    -------
    bool
        DESCRIPTION.

    """
    try:
        # file count per SSRUGO version 2.3.3    
        mainN = 5 # template database may not have been requested and is not needed
        spatN = 26
        tabN = 68
        spatF = os.path.join(newFolder, 'spatial')
        tabF = os.path.join(newFolder, 'tabular')
        saCatalog = os.path.join(tabF, "sacatlog.txt")
        
        dbDate = 0
        surveyDate = surveyDate.replace('-', '')
        
        # Check folders for completeness
        if len(os.listdir(newFolder)) >= mainN:
            if os.path.isdir(spatF) and len(os.listdir(spatF)) >= spatN:
                if os.path.isdir(tabF) and len(os.listdir(tabF)) >= tabN:
                    if os.path.isfile(saCatalog):
                        fh = open(saCatalog, "r")
                        rec = fh.readline()
                        fh.close()
                        # Example date (which is index 3 in pipe-delimited file):  9/23/2014 6:49:27
                        vals = rec.split("|")
                        recDate = vals[3]
                        wssDate = "%m/%d/%Y %H:%M:%S"  # string date format used for SAVEREST in text file
                        intDate = "%Y%m%d"             # YYYYMMDD format for comparison
                        dateObj = datetime.strptime(recDate, wssDate)
                        dbDate = int(dateObj.strftime(intDate))
                        
                        if surveyDate <= dbDate:
                            # download_b = False
                            msgs = f"\nLocal dataset for {areaSym} already exists (date of {dbDate})"
                            return (1, msgs)
        # Current version is incomplete, remove and flag to download
        shutil.rmtree(newFolder, True)
        if os.path.isdir(newFolder):
            try:
                shutil.rmtree(newFolder)
            except PermissionError as e:
                msgs = f"Failed to delete obsolete dataset ({newFolder})"
                msgs = msgs + f'\nPermission Error: {e}'
                return [1, msgs]
        else:
            return (True, surveyDate)
        if os.path.isdir(newFolder): 
           msgs = f"Failed to delete obsolete dataset ({newFolder})"
           return (1, msgs)
        else: 
           return (0, surveyDate)

    except arcpy.ExecuteError:
        func = sys._getframe(  ).f_code.co_name
        msgs = arcpyErr(func)
        return [2, msgs]
    except:
        func = sys._getframe(  ).f_code.co_name
        msgs = pyErr(func)
        return [2, msgs]



## ===================================================================================
def GetDownload(areaSym, surveyDate, outFolder):
    """download survey from Web Soil Survey URL    
    Only the version of zip file without a Template database is downloaded. The user
    must have a locale copy of the Template database that has been modified to allow
    automatic tabular imports.
    

    Parameters
    ----------
    areasym : str
        The area symbol of current soil survey area being processed.
    surveyDate : TYPE
        DESCRIPTION.
    newFolder : TYPE
        DESCRIPTION.
    outFolder : TYPE
        DESCRIPTION.

    Returns
    -------
    bool
        Successfull download.

    """

    # Create a custom logger
    #logger = logging.getLogger(__name__)
    #logger.disabled = False
    # Create handlers
    #f_handler = logging.FileHandler('fuckyou.log')
    #f_handler.setLevel(logging.ERROR)
    # Create formatters and add it to handlers
    #f_format = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    #f_handler.setFormatter(f_format)
    # Add handlers to the logger
    #logger.addHandler(f_handler)
    #logger.warning('This is a warning')
    #logger.error('This is an error')


    
    # download survey from Web Soil Survey URL and return name of the zip file
    # want to set this up so that download will retry several times in case of error
    # return empty string in case of complete failure. Allow main to skip a failed
    # survey, but keep a list of failures
    #
    # Only the version of zip file without a Template database is downloaded. The user
    # must have a locale copy of the Template database that has been modified to allow
    # automatic tabular imports.

    # create URL string from survey string and WSS 3.0 cache URL
    baseURL = "https://websoilsurvey.sc.egov.usda.gov/DSD/Download/Cache/SSA/"

    try: 
        # Use this zipfile for downloads without the Template database
        # date.fromisoformat(surveyDate)
        # zipDate = str(surveyDate)[0:4] + "-" + str(surveyDate)[4:6] + "-" + str(surveyDate)[6:8]
        zipName = f"wss_SSA_{areaSym}_[{surveyDate}].zip"

        # Use this URL for downloads with the state or US_2003 database
        #zipName = "wss_SSA_" + areaSym + db + "_[" + surveyDate + "].zip"

        zipURL = baseURL + zipName
        arcpy.AddMessage(f"\tDownloading survey {areaSym} from Web Soil Survey...")

        # Open request to Web Soil Survey for that zip file
        r = requests.get(zipURL)
        # if r.status_code>=400:
        z = zipfile.ZipFile(io.BytesIO(r.content))
        z.extractall(path=outFolder)
        
        # ---- Check naming
        # rename output folder to NRCS Geodata Standard for Soils
        # if os.path.isdir(f"{outFolder}/{areaSym.upper()}"):
        #     # this must be a newer zip file using the uppercase AREASYMBOL directory
        #     os.rename(f"{outFolder}/{areaSym.upper()}", newFolder)
            
        # elif os.path.isdir(f"{outFolder}/{zipName[:-4]}"):
        #     # this is an older zip file that has the 'wss_' directory structure
        #     os.rename(f"{outFolder}/{zipName[:-4]}", newFolder)
        
        # if os.path.isdir(newFolder):
        #     # this is a future zip file using the correct field office naming convention (soil_ne109)
        #     # it does not require renaming.
        #     pass
        
        if not os.path.isdir(outFolder):
            # none of the subfolders within the zip file match any of the expected names
            # msgs = "Subfolder within the zip file does not match any of the standard names"
            msgs = "File did not appear to unzip successfuly"
            return [1, msgs]
        return [0, None]

    except requests.HTTPError as e:
        msgs = f'HTTP Error {e}'
        return [2, msgs]

    except requests.Timeout:
        msgs = 'Soil Data Access timeout error'
        return [2, msgs]

    except socket.error as e:
        msgs = f'Socket error: {e}'
        msgs = msgs + '\nAlso possible File Explorer needs tob be closed'
        return [2, msgs]
    
    except socket.timeout as e:
        msgs = f'Socket timeout error: {e}'
        return [2, msgs]

    except URLError as e:
        msgs = f'URL error: {e}'
        msgs = msgs + f"\n{zipURL}"
        return [2, msgs]

    except HTTPError as e:
        msgs = f'HTTP error: {e}'
        msgs = msgs + f"\n{zipURL}"
        return [2, msgs]
    
    except zipfile.BadZipfile:
        msgs = "Bad zip file?"
        msgs = msgs + f"\n{zipURL}"
        return [2, msgs]
    
    except arcpy.ExecuteError:
        func = sys._getframe(  ).f_code.co_name
        msgs = arcpyErr(func)
        return [2, msgs]
    
    except:
        func = sys._getframe(  ).f_code.co_name
        msgs = pyErr(func)
        return [2, msgs]



## ===================================================================================
def ProcessSurvey(outputFolder, areaSym, surveyInfo, overwriteFlg):
    """Manages the download process for each soil survey
    

    Parameters
    ----------
    outputFolder : str
        Folder location for the downloaded SSURGO datasets.

    areaSym : str
        The area symbol of current soil survey area being processed.
    surveyInfo : list
        The date the soil survey area was updated on WSS and survey name


    Returns
    -------
    str
        keywords: 'Successful', 'Skipped',or 'Failed'.

    """
    # Download and import the specified SSURGO dataset

    try:
        # get date string
        surveyDate = surveyInfo[0].strip()
        # get survey name
        # set standard final path and name for template database
        # newFolder = f"{outputFolder}/soil_{areaSym.lower()}"
        # ---- Call CheckExistingDataset
        if not surveyDate:
            msgs = "No Survey Date in WSS SSA label"
            return [1, msgs]
        if overwriteFlg == False:
            if os.path.isdir(os.path.join(outputFolder,areaSym)):
                msgs = 'Will not overwrite ' + areaSym
                return [0, msgs]
        # if os.path.isdir(newFolder):
        #     cue, msgs = CheckExistingDataset(areaSym, surveyDate, newFolder)
        #     if cue:
        #         return [cue, areaSym, msgs]
        #     else:
        #         download_b = True
        # else:
        #     download_b = True
        # ---- Call GetDownload
        # First attempt to download zip file
        # if download_b:
            # Does it need to specify download with .mdb file?
        dcue, msg = GetDownload(areaSym, surveyDate, outputFolder)
        if dcue:
            # Try downloading zip file a second time
            sleep(5)
            dcue, msgs = GetDownload(areaSym, surveyDate, outputFolder)
            if dcue:
                # Failed to download
                return [dcue, msgs]
        msgs = '\tSurvey successfully downloaded'
        return [0, msgs]

    except arcpy.ExecuteError:
        func = sys._getframe(  ).f_code.co_name
        msgs = arcpyErr(func)
        return [2, msgs]
    
    except:
        func = sys._getframe(  ).f_code.co_name
        msgs = pyErr(func)
        return [2, msgs]


# %% Main
if __name__ == '__main__':

    try:
        v = 0.4
        arcpy.AddMessage(f"version {v}")
        # arcpy.env.parallelProcessingFactor = "75%"
        pCores = 12
        arcpy.env.overwriteOutput = True

        # ---- Parameters
        outputFolder = arcpy.GetParameterAsText(1)
        surveyList = arcpy.GetParameter(2)
        overwriteFlg = arcpy.GetParameter(3)

        # ---- Setup

        arcpy.AddMessage(f"\n{len(surveyList)} soil survey(s) selected for Web Soil Survey download")

        # set workspace to output folder
        env.workspace = outputFolder
        # ---- Prime For-loop
        # Create ordered list by Areasymbol
        # AREASYMBOL: Date, Survey Name, State
        paramSet = [{'areaSym': s.split(',')[0].strip().upper(),
                   'surveyInfo': s.split(',')[1:]} 
                  for s in surveyList
                  if s.split(',')[0].strip().upper() != 'HT600']
        
        # %%% By Area Symbol
        surveyCount = len(paramSet)
        threadCount = min(mp.cpu_count(), surveyCount)
        arcpy.AddMessage(f"\tRunning on {threadCount} threads.\n")
        successCount = 0
        failList = []
        # Run import process
        # ---- Call ProcessSurvey
        constSet = {'outputFolder': outputFolder,'overwriteFlg':overwriteFlg}
        for paramBack, output in concurrently(ProcessSurvey,
                                             threadCount,
                                             paramSet,
                                             constSet):
            try:
                outcome, msg = output
                # arcpy.AddMessage(f"{msg}")
                # arcpy.AddMessage(f"{outcome}: {paramBack}")
                # ssa = paramBack[0]
                if outcome == 0:
                    arcpy.AddMessage(f"{paramBack['areaSym']}:\t\n{msg}")
                    successCount += 1
                elif outcome == 1:
                    # arcpy.AddWarning(f"{ssa}:\n{msg}")
                    arcpy.AddWarning(f"{paramBack['areaSym']}:\t\n{msg}")
                    failList.append(paramBack['areaSym'])
                else:
                    # arcpy.AddError(f"{ssa}:\n{msg}")
                    arcpy.AddError(f"{paramBack['areaSym']}:\t\n{msg}")
                    failList.append(paramBack['areaSym'])
            except GeneratorExit:
                pass
                       
        arcpy.SetProgressorLabel("Processing complete...")
        arcpy.AddMessage(f"\nSuccessfully downloaded {successCount} of {surveyCount} surveys.")
        if failList:
            arcpy.AddWarning(f"\n{len(failList)} surveys failed to load:")
            for ssa in failList:
                arcpy.AddWarning(f"\t{ssa}")

    except arcpy.ExecuteError:
        func = 'Main'
        arcpyErr(func)
    
    except:
        func = 'Main'
        pyErr(func)

